# Copyright (c) 2022, Voltron Data.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import textwrap
from dataclasses import dataclass
from typing import List

from .clients import CheckStatus
from .talk_to_conbench import RunComparison


def _clean(text: str) -> str:
    """Clean text so it displays nicely as GitHub Markdown."""
    return textwrap.fill(textwrap.dedent(text), 10000).replace("  ", "\n\n").strip()


@dataclass
class _CaseInfo:
    run_id: str
    run_reason: str
    run_time: str
    run_link: str
    case_name: str
    case_link: str

    @property
    def Run_Reason(self):
        return self.run_reason.title()


def _list_cases(case_infos: List[_CaseInfo]) -> str:
    """Create a Markdown list of case information."""
    out = ""
    previous_run_id = ""

    for case in case_infos:
        if case.run_id != previous_run_id:
            out += f"\n\n- {case.Run_Reason} Run at [{case.run_time}]({case.run_link})"
            previous_run_id = case.run_id
        out += f"\n  - [{case.case_name}]({case.case_link})"

    if out:
        out += "\n\n"

    return out


def benchmarks_with_errors(comparisons: List[RunComparison]) -> List[_CaseInfo]:
    """Find information about benchmark cases that had errors."""
    out = []

    for comparison in comparisons:
        if comparison.compare_results:
            out += [
                _CaseInfo(
                    run_id=comparison.contender_id,
                    run_reason=comparison.contender_reason,
                    run_time=comparison.contender_datetime,
                    run_link=comparison.compare_link,
                    case_name=case["benchmark"],
                    case_link=comparison.case_link(case["contender_id"]),
                )
                for case in comparison.compare_results
                if case["contender_error"]
            ]
        elif comparison.benchmark_results:
            out += [
                _CaseInfo(
                    run_id=comparison.contender_id,
                    run_reason=comparison.contender_reason,
                    run_time=comparison.contender_datetime,
                    run_link=comparison.contender_link,
                    case_name=case["tags"].get("name", str(case["tags"])),
                    case_link=comparison.case_link(case["id"]),
                )
                for case in comparison.benchmark_results
                if case["error"]
            ]

    return out


def benchmarks_with_z_regressions(comparisons: List[RunComparison]) -> List[_CaseInfo]:
    """Find information about benchmark cases whose z-scores were extreme enough to
    constitute a regression.
    """
    out = []

    for comparison in comparisons:
        if comparison.compare_results:
            out += [
                _CaseInfo(
                    run_id=comparison.contender_id,
                    run_reason=comparison.contender_reason,
                    run_time=comparison.contender_datetime,
                    run_link=comparison.compare_link,
                    case_name=case["benchmark"],
                    case_link=comparison.case_link(case["contender_id"]),
                )
                for case in comparison.compare_results
                if case["contender_z_regression"]
            ]

    return out


def regression_summary(
    comparisons: List[RunComparison], warn_if_baseline_isnt_parent: bool
) -> str:
    """Generate a Markdown summary of what happened regarding errors and regressions."""
    sha = comparisons[0].contender_info["commit"]["sha"][:8]
    errors = benchmarks_with_errors(comparisons)
    regressions = benchmarks_with_z_regressions(comparisons)
    summary = ""

    if errors:
        summary += _clean(
            """
            ## Benchmarks with errors

            These are errors that were caught while running the benchmarks. You can
            click the link next to each case to go to the Conbench entry for that
            benchmark, which might have more information about what the error was.
            """
        )
        summary += _list_cases(errors)

    summary += "## Benchmarks with performance regressions\n\n"

    if not any(comparison.baseline_info for comparison in comparisons):
        summary += _clean(
            f"""
            Conbench could not find a baseline run for contender commit `{sha}`. A
            baseline run needs to be on the default branch in the same repository, with
            the same hardware and context, and have at least one of the same benchmark
            cases.
            """
        )
        return summary

    summary += _clean(
        f"""
        Contender commit `{sha}` had {len(regressions)} performance regression(s)
        compared to its baseline commit.
        """
    )
    summary += "\n\n"

    if regressions:
        summary += "### Benchmarks with regressions:"
        summary += _list_cases(regressions)

    if (
        any(not comparison.baseline_is_parent for comparison in comparisons)
        and warn_if_baseline_isnt_parent
    ):
        summary += _clean(
            """
            ### Note

            The baseline commit was not the immediate parent of the contender commit.
            See the link below for details.
            """
        )

    return summary


def regression_details(comparisons: List[RunComparison]) -> str:
    """Generate Markdown details of what happened regarding regressions."""
    if not any(comparison.baseline_info for comparison in comparisons):
        return None

    z_score_threshold = comparisons[0].compare_results[0]["threshold_z"]
    details = _clean(
        f"""
        Conbench has details about {len(comparisons)} total run(s) on this commit.

        This report was generated using a z-score threshold of {z_score_threshold}. A
        regression is defined as a benchmark exhibiting a z-score higher than the
        threshold in the "bad" direction (e.g. down for iterations per second; up for
        total time taken).
        """
    )
    return details


def regression_check_status(
    comparisons: List[RunComparison],
) -> CheckStatus:
    """Return a different status based on errors and regressions."""
    regressions = benchmarks_with_z_regressions(comparisons)

    if any(comparison.has_errors for comparison in comparisons):
        # has errors
        return CheckStatus.ACTION_REQUIRED
    if not any(comparison.baseline_info for comparison in comparisons):
        # no baseline runs found
        return CheckStatus.SKIPPED
    elif regressions:
        # at least one regression
        return CheckStatus.FAILURE
    else:
        # no regressions
        return CheckStatus.SUCCESS



# Copyright (c) 2022, Voltron Data.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import abc
import datetime
import enum
import os
import textwrap
from json import dumps
from typing import Optional

import jwt
import requests
from requests.adapters import HTTPAdapter
from urllib3.util.retry import Retry

from .log import fatal_and_log, log


class _BaseClient(abc.ABC):
    """A client to interact with an API.

    Parameters
    ----------
    adapter
        A requests adapter to mount to the requests session. If not given, one will be
        created with a backoff retry strategy.
    """

    base_url: str
    timeout_s = 10

    def __init__(self, adapter: Optional[HTTPAdapter]):
        if not adapter:
            retry_strategy = Retry(
                total=5,
                status_forcelist=frozenset((429, 502, 503, 504)),
                backoff_factor=4,  # will retry in 2, 4, 8, 16, 32 seconds
            )
            adapter = HTTPAdapter(max_retries=retry_strategy)

        self.session = requests.Session()
        self.session.mount("https://", adapter)

    def get(self, path: str, params: Optional[dict] = None) -> dict:
        url = self.base_url + path
        log.debug(f"GET {url} {params=}")
        res = self.session.get(url=url, params=params, timeout=self.timeout_s)
        self._maybe_raise(res)
        return res.json()

    def post(self, path: str, json: dict = None) -> Optional[dict]:
        json = json or {}
        url = self.base_url + path
        log.debug(f"POST {url} {dumps(json)}")
        res = self.session.post(url=url, json=json, timeout=self.timeout_s)
        self._maybe_raise(res)
        if res.content:
            return res.json()

    @staticmethod
    def _maybe_raise(res: requests.Response):
        try:
            res.raise_for_status()
        except requests.HTTPError as e:
            try:
                res_content = e.response.content.decode()
            except AttributeError:
                res_content = e.response.content
            log.error(f"Response content: {res_content}")
            raise


class GitHubAppClient(_BaseClient):
    """A client to interact with a GitHub App.

    Parameters
    ----------
    adapter
        A requests adapter to mount to the requests session. If not given, one will be
        created with a backoff retry strategy.

    Environment variables
    ---------------------
    GITHUB_APP_ID
        The numeric GitHub App ID you can get from its settings page.
    GITHUB_APP_PRIVATE_KEY
        The full contents of the private key file downloaded from the App's settings
        page.
    """

    def __init__(self, adapter: Optional[HTTPAdapter] = None):
        app_id = os.getenv("GITHUB_APP_ID")
        if not app_id:
            fatal_and_log("Environment variable GITHUB_APP_ID not found")

        private_key = os.getenv("GITHUB_APP_PRIVATE_KEY")
        if not private_key:
            fatal_and_log("Environment variable GITHUB_APP_PRIVATE_KEY not found")

        super().__init__(adapter=adapter)
        encoded_jwt = self._encode_jwt(app_id=app_id, private_key=private_key)
        self.session.headers = {"Authorization": f"Bearer {encoded_jwt}"}
        self.base_url = "https://api.github.com/app"

    @staticmethod
    def _encode_jwt(app_id: str, private_key: str) -> str:
        """Create, sign, and encode a JSON web token to use for GitHub App endpoints."""
        payload = {
            "iss": app_id,
            "iat": datetime.datetime.utcnow() - datetime.timedelta(minutes=1),
            "exp": datetime.datetime.utcnow() + datetime.timedelta(minutes=10),
        }
        encoded_jwt = jwt.encode(payload=payload, key=private_key, algorithm="RS256")
        return encoded_jwt

    def get_app_access_token(self) -> str:
        """Authenticate as the GitHub App and generate an app installation access token.

        The token lasts for 1 hour, which should be plenty of time to do anything this
        package needs to do.

        Returns
        -------
        str
            A temporary API token to use for the GitHub API endpoints that the app has
            permission to access.
        """
        # We tell developers to create a new app for each organization, so they can
        # control the private key. So there should be exactly 1 installation here.
        # (Note: 1 installation could have multiple repos in the same organization.)
        installations = self.get("/installations")
        install_id = installations[0]["id"]

        token_info = self.post(f"/installations/{install_id}/access_tokens")
        return token_info["token"]


# used as inputs to some GitHubRepoClient methods
class StatusState(str, enum.Enum):
    ERROR = "error"
    FAILURE = "failure"
    PENDING = "pending"
    SUCCESS = "success"


# used as inputs to some GitHubRepoClient methods
class CheckStatus(str, enum.Enum):
    # statuses
    QUEUED = "queued"
    IN_PROGRESS = "in_progress"
    # conclusions
    ACTION_REQUIRED = "action_required"
    CANCELLED = "cancelled"
    FAILURE = "failure"
    NEUTRAL = "neutral"
    SUCCESS = "success"
    SKIPPED = "skipped"
    TIMED_OUT = "timed_out"


class GitHubRepoClient(_BaseClient):
    """A client to interact with a GitHub repo.

    You may authenticate with the GitHub API using a GitHub Personal Access Token or a
    GitHub App. The correct environment variables must be set depending on which method
    of authentication you're using. If all are set, the App method will be used.

    Parameters
    ----------
    repo
        The repo name, in the form 'owner/repo'.
    adapter
        A requests adapter to mount to the requests session. If not given, one will be
        created with a backoff retry strategy.

    Environment variables
    ---------------------
    GITHUB_APP_ID
        The numeric GitHub App ID you can get from its settings page. Only used for
        GitHub App authentication.
    GITHUB_APP_PRIVATE_KEY
        The full contents of the private key file downloaded from the App's settings
        page. Only used for GitHub App authentication.
    GITHUB_API_TOKEN
        A GitHub API token with ``repo`` access. Only used for Personal Access Token
        authentication.
    """

    def __init__(self, repo: str, adapter: Optional[HTTPAdapter] = None):
        if os.getenv("GITHUB_APP_ID") or os.getenv("GITHUB_APP_PRIVATE_KEY"):
            log.info("Attempting to authenticate as a GitHub App.")
            app_client = GitHubAppClient(adapter=adapter)
            token = app_client.get_app_access_token()
        else:
            token = os.getenv("GITHUB_API_TOKEN")
            if not token:
                fatal_and_log("Environment variable GITHUB_API_TOKEN not found.")

        super().__init__(adapter=adapter)
        self.session.headers = {"Authorization": f"Bearer {token}"}
        self.base_url = f"https://api.github.com/repos/{repo}"

    def create_pull_request_comment(
        self,
        comment: str,
        *,
        pull_number: Optional[int] = None,
        commit_sha: Optional[str] = None,
    ):
        """Create a comment on a pull request, specified either by pull request number
        or commit SHA.

        Parameters
        ----------
        comment
            The comment text.
        pull_number
            The number of the pull request. Specify either this or ``commit_sha``.
        commit_sha
            The SHA of a commit associated with the pull request. Specify either this
            or ``pull_number``.
        """
        if not pull_number and not commit_sha:
            fatal_and_log("pull_number and commit_sha are both missing")

        if commit_sha:
            pull_numbers = [
                pull["number"] for pull in self.get(f"/commits/{commit_sha}/pulls")
            ]
            if len(pull_numbers) != 1:
                fatal_and_log(
                    "Need exactly 1 pull request associated with commit "
                    f"'{commit_sha}'. Found {pull_numbers}."
                )
            pull_number = pull_numbers[0]

        log.info(
            f"Posting the following message to pull request #{pull_number}:\n\n"
            + comment
        )
        return self.post(f"/issues/{pull_number}/comments", json={"body": comment})

    def update_commit_status(
        self,
        commit_sha: str,
        title: str,
        description: str,
        state: StatusState,
        details_url: Optional[str] = None,
    ) -> dict:
        """Update the GitHub status of a commit.

        A commit may have many statuses, each with their own title. Updating a previous
        status with the same title for a given commit will result in overwriting that
        status on that commit.

        Parameters
        ----------
        commit_sha
            The 40-character SHA of the commit to update.
        title
            The title of the status. Subsequent updates with the same title will update
            the same status.
        description
            The short description of the status.
        state
            The overall status of the commit. Must be one of the StatusState enum
            values.
        details_url
            A URL to be linked to when clicking on status Details. Default None.

        Returns
        -------
        dict
            GitHub's details about the new status.
        """
        if not isinstance(state, StatusState):
            fatal_and_log("state must be a StatusState", etype=TypeError)

        json = {
            "state": state.value,
            "description": textwrap.shorten(description, 140),
            "context": title,
        }
        if details_url:
            json["target_url"] = details_url

        return self.post(f"/statuses/{commit_sha}", json=json)

    def update_check(
        self,
        name: str,
        commit_sha: str,
        status: CheckStatus,
        title: Optional[str] = None,
        summary: Optional[str] = None,
        details: Optional[str] = None,
        details_url: Optional[str] = None,
    ):
        """Adds a new (or updates an existing) GitHub Check on a commit.

        A GitHub Check is a more fully-featured commit status, but you must be
        authenticated as a GitHub App to update checks. Subsequent uses of this method
        with the same ``name`` will overwrite the previous check.

        Parameters
        ----------
        name
            The name of the check. Subsequent updates with the same name will overwrite
            the previous check.
        commit_sha
            The 40-character SHA of the commit to update.
        status
            The overall check status. Must be one of the CheckStatus enum values. If
            it's QUEUED or IN_PROGRESS, the "started_at" field will be sent in the
            payload with the current time in UTC. If it's another value, the
            "completed_at" field will be sent instead.
        title
            The short title of the check results. Default None. If supplied, summary
            must be supplied.
        summary
            A longer summary of the check results. Supports Markdown. Default None. If
            supplied, title must be supplied.
        details
            Details about the check results. Supports Markdown. Default None.
        details_url
            A URL to be linked to when clicking on check Details. Default None.

        Returns
        -------
        dict
            GitHub's details about the new status.
        """
        json = {"name": name, "head_sha": commit_sha}

        if status in [CheckStatus.QUEUED, CheckStatus.IN_PROGRESS]:
            json["status"] = status.value
            json["started_at"] = datetime.datetime.utcnow().isoformat() + "Z"
        elif isinstance(status, CheckStatus):
            json["conclusion"] = status.value
            json["completed_at"] = datetime.datetime.utcnow().isoformat() + "Z"
        else:
            fatal_and_log("status must be a CheckStatus", etype=TypeError)

        if title:
            json["output"] = {"title": title, "summary": summary}
            if details:
                json["output"]["text"] = details

        if details_url:
            json["details_url"] = details_url

        return self.post("/check-runs", json=json)


class ConbenchClient(_BaseClient):
    """A client to interact with a Conbench server.

    Parameters
    ----------
    adapter
        A requests adapter to mount to the requests session. If not given, one will be
        created with a backoff retry strategy.

    Environment variables
    ---------------------
    CONBENCH_URL
        The URL of the Conbench server. Required.
    CONBENCH_EMAIL
        The email to use for Conbench login. Only required if the server is private.
    CONBENCH_PASSWORD
        The password to use for Conbench login. Only required if the server is private.
    """

    def __init__(self, adapter: Optional[HTTPAdapter] = None):
        url = os.getenv("CONBENCH_URL")
        if not url:
            fatal_and_log("Environment variable CONBENCH_URL not found")

        super().__init__(adapter=adapter)
        self.base_url = url + "/api"

        login_creds = {
            "email": os.getenv("CONBENCH_EMAIL"),
            "password": os.getenv("CONBENCH_PASSWORD"),
        }
        if login_creds["email"] and login_creds["password"]:
            self.post("/login/", json=login_creds)



# Copyright (c) 2022, Voltron Data.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from dataclasses import dataclass
from typing import List, Optional

from .clients import ConbenchClient
from .log import fatal_and_log, log


@dataclass
class RunComparison:
    """Track info about a comparison between a contender run and its baseline. Used for
    outputting information from get_comparison_to_baseline().

    Parameters
    ----------
    contender_info
        The dict returned from Conbench when hitting /runs/{contender_run_id}. Contains
        info about the run's ID, commit, errors, links, etc.
    baseline_info
        The dict returned from Conbench when hitting /runs/{baseline_run_id}, if a
        baseline run exists for this contender run. Contains info about the run's ID,
        commit, errors, links, etc.
    compare_results
        The list returned from Conbench when hitting
        /compare/runs/{baseline_run_id}...{contender_run_id}, if a baseline run exists
        for this contender run. Contains a comparison for every case run to its
        baseline, including the statistics and regression analysis.
    benchmark_results
        The list returned from Conbench when hitting
        /benchmarks?run_id={contender_run_id}, if the contender run has errors. Contains
        info about each case in the contender run, including statistics and tracebacks.
        Only used when a baseline run doesn't exist, because otherwise all this
        information is already in the compare_results.
    """

    contender_info: dict
    baseline_info: Optional[dict] = None
    compare_results: Optional[List[dict]] = None
    benchmark_results: Optional[List[dict]] = None

    @property
    def baseline_is_parent(self) -> Optional[bool]:
        """Whether the baseline run is on a commit that's the immediate parent of the
        contender commit.
        """
        if self.baseline_info:
            return (
                self.baseline_info["commit"]["sha"]
                == self.contender_info["commit"]["parent_sha"]
            )

    @property
    def contender_reason(self) -> str:
        """The contender run reason."""
        return self.contender_info["reason"]

    @property
    def contender_datetime(self) -> str:
        """The contender run datetime."""
        dt: str = self.contender_info["timestamp"]
        return dt.replace("T", " ")

    @property
    def contender_link(self) -> str:
        """The link to the contender run page in the webapp."""
        return f"{self._app_url}/runs/{self.contender_id}"

    @property
    def compare_link(self) -> Optional[str]:
        """The link to the run comparison page in the webapp."""
        if self._compare_path:
            # self._compare_path has a leading slash already
            return f"{self._app_url}{self._compare_path}"

    def case_link(self, case_id: str) -> str:
        """Get the link to a specific benchmark case result in the webapp."""
        return f"{self._app_url}/benchmarks/{case_id}"

    @property
    def has_errors(self) -> bool:
        """Whether this run has any benchmark errors."""
        return self.contender_info["has_errors"]

    @property
    def contender_id(self) -> str:
        """The contender run_id."""
        return self.contender_info["id"]

    @property
    def _baseline_id(self) -> Optional[str]:
        """The baseline run_id."""
        if self.baseline_info:
            return self.baseline_info["id"]

    @property
    def _app_url(self) -> str:
        """The base URL to use for links to the webapp."""
        self_link: str = self.contender_info["links"]["self"]
        return self_link.rsplit("/api/", 1)[0]

    @property
    def _compare_path(self) -> Optional[str]:
        """The API path to get comparisons between the baseline and contender."""
        if self._baseline_id:
            return f"/compare/runs/{self._baseline_id}...{self.contender_id}/"

    @property
    def _baseline_path(self) -> Optional[str]:
        """The API path to get the baseline info."""
        baseline_link: Optional[str] = self.contender_info["links"].get("baseline")
        if baseline_link:
            return baseline_link.rsplit("/api", 1)[-1]


def get_comparison_to_baseline(
    conbench: ConbenchClient,
    contender_sha: str,
    z_score_threshold: Optional[float] = None,
) -> List[RunComparison]:
    """Get benchmark comparisons between the given contender commit and its baseline
    commit.

    The baseline commit is defined by conbench, and it's typically the most recent
    ancestor of the contender commit that's on the default branch.

    Parameters
    ----------
    conbench
        A ConbenchClient instance.
    contender_sha
        The commit SHA of the contender commit to compare. Needs to match EXACTLY what
        conbench has stored; typically 40 characters. It can't be a shortened version of
        the SHA.
    z_score_threshold
        The (positive) z-score threshold to send to the conbench compare endpoint.
        Benchmarks with a z-score more extreme than this threshold will be marked as
        regressions or improvements in the result. Default is to use whatever conbench
        uses for default.

    Returns
    -------
    List[RunComparison]
        Information about each run associated with the contender commit, and a
        comparison to its baseline run if that exists.
    """
    out_list = []
    contender_run_ids = [
        run["id"] for run in conbench.get("/runs/", params={"sha": contender_sha})
    ]
    if not contender_run_ids:
        fatal_and_log(
            f"Contender commit '{contender_sha}' doesn't have any runs in conbench."
        )

    log.info(f"Getting comparisons from {len(contender_run_ids)} run(s)")
    for run_id in contender_run_ids:
        run_comparison = RunComparison(contender_info=conbench.get(f"/runs/{run_id}/"))

        if run_comparison._baseline_path:
            run_comparison.baseline_info = conbench.get(run_comparison._baseline_path)

            compare_params = (
                {"threshold_z": z_score_threshold} if z_score_threshold else None
            )
            run_comparison.compare_results = conbench.get(
                run_comparison._compare_path, params=compare_params
            )

        else:
            log.warning(
                "Conbench could not find a baseline run for the contender run "
                f"{run_id}. A baseline run needs to be on the default branch in the "
                "same repository, with the same hardware and context, and have at "
                "least one of the same benchmark cases."
            )
            if run_comparison.has_errors:
                # get more information so we have more details about errors
                run_comparison.benchmark_results = conbench.get(
                    "/benchmarks/", params={"run_id": run_id}
                )

        out_list.append(run_comparison)

    return out_list



# Copyright (c) 2022, Voltron Data.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
from typing import Optional

from .clients import CheckStatus, ConbenchClient, GitHubRepoClient, StatusState
from .log import log
from .parse_conbench import (
    _clean,
    benchmarks_with_z_regressions,
    regression_check_status,
    regression_details,
    regression_summary,
)
from .talk_to_conbench import get_comparison_to_baseline


def update_github_status_based_on_regressions(
    contender_sha: str,
    z_score_threshold: Optional[float] = None,
    repo: Optional[str] = None,
    github: Optional[GitHubRepoClient] = None,
    conbench: Optional[ConbenchClient] = None,
) -> dict:
    """Grab the benchmark result comparisons for a given contender commit, and post to
    GitHub whether there were any regressions, in the form of a commit status.

    Parameters
    ----------
    contender_sha
        The SHA of the contender commit to compare. Needs to match EXACTLY what
        conbench has stored; typically 40 characters. It can't be a shortened
        version of the SHA.
    z_score_threshold
        The (positive) z-score threshold. Benchmarks with a z-score more extreme than
        this threshold will be marked as regressions. Default is to use whatever
        conbench uses for default.
    repo
        The repo name to post the status to, in the form 'owner/repo'. Either provide
        this or ``github``.
    github
        A GitHubRepoClient instance. Either provide this or ``repo``.
    conbench
        A ConbenchClient instance. If not given, one will be created using the standard
        environment variables.

    Environment variables
    ---------------------
    BUILD_URL
        The URL of the build running this code. If provided, the GitHub status will link
        to the build when there's an error in this workflow.
    GITHUB_APP_ID
        The ID of a GitHub App that has been set up according to this package's
        instructions and installed to your repo. Recommended over GITHUB_API_TOKEN. Only
        required if a GitHubRepoClient is not provided.
    GITHUB_APP_PRIVATE_KEY
        The private key file contents of a GitHub App that has been set up according to
        this package's instructions and installed to your repo. Recommended over
        GITHUB_API_TOKEN. Only required if GitHubRepoClient is not provided.
    GITHUB_API_TOKEN
        A GitHub Personal Access Token with the ``repo:status`` permission. Only
        required if not going with GitHub App authentication and if a GitHubRepoClient
        is not provided.
    CONBENCH_URL
        The URL of the Conbench server. Only required if a ConbenchClient is not
        provided.
    CONBENCH_EMAIL
        The email to use for Conbench login. Only required if a ConbenchClient is not
        provided and the server is private.
    CONBENCH_PASSWORD
        The password to use for Conbench login. Only required if a ConbenchClient is not
        provided and the server is private.

    Returns
    -------
    dict
        GitHub's details about the new status.
    """
    build_url = os.getenv("BUILD_URL")
    github = github or GitHubRepoClient(repo=repo)

    def update_status(description, state, details_url):
        """Shortcut for updating the "conbench" status on the given SHA, with debug
        logging.
        """
        res = github.update_commit_status(
            commit_sha=contender_sha,
            title="conbench",
            description=description,
            state=state,
            details_url=details_url,
        )
        log.debug(res)
        return res

    # mark the task as pending
    update_status(
        description="Finding possible regressions",
        state=StatusState.PENDING,
        details_url=build_url,
    )

    # If anything above this line fails, we can't tell github that it failed.
    # If anything in here fails, we can!
    try:
        conbench = conbench or ConbenchClient()
        comparisons = get_comparison_to_baseline(
            conbench, contender_sha, z_score_threshold
        )
        regressions = benchmarks_with_z_regressions(comparisons)
        log.info(f"Found the following regressions: {regressions}")

        if not any(comparison.baseline_info for comparison in comparisons):
            desc = "Could not find any baseline runs to compare to"
            state = StatusState.SUCCESS
        elif regressions:
            desc = f"There were {len(regressions)} benchmark regressions in this commit"
            state = StatusState.FAILURE
        else:
            desc = "There were no benchmark regressions in this commit"
            state = StatusState.SUCCESS

        # point to the homepage table filtered to runs of this commit
        url = f"{os.environ['CONBENCH_URL']}/?search={contender_sha}"
        return update_status(description=desc, state=state, details_url=url)

    except Exception as e:
        update_status(
            description=f"Failed finding regressions: {e}",
            state=StatusState.ERROR,
            details_url=build_url,
        )
        log.error(f"Updated status with error: {e}")
        raise


def update_github_check_based_on_regressions(
    contender_sha: str,
    z_score_threshold: Optional[float] = None,
    warn_if_baseline_isnt_parent: bool = True,
    repo: Optional[str] = None,
    github: Optional[GitHubRepoClient] = None,
    conbench: Optional[ConbenchClient] = None,
) -> dict:
    """Grab the benchmark result comparisons for a given contender commit, and post to
    GitHub whether there were any regressions, in the form of a commit check.

    You must use GitHub App authentication to use this workflow.

    Parameters
    ----------
    contender_sha
        The SHA of the contender commit to compare. Needs to match EXACTLY what
        conbench has stored; typically 40 characters. It can't be a shortened
        version of the SHA.
    z_score_threshold
        The (positive) z-score threshold. Benchmarks with a z-score more extreme than
        this threshold will be marked as regressions. Default is to use whatever
        conbench uses for default.
    warn_if_baseline_isnt_parent
        If True, will add a warning to all reports generated where the baseline commit
        isn't the contender commit's direct parent. This is good to leave True for
        workflows run on the default branch, but might be noisy for workflows run on
        pull request commits.
    repo
        The repo name to post the status to, in the form 'owner/repo'. Either provide
        this or ``github``.
    github
        A GitHubRepoClient instance. Either provide this or ``repo``.
    conbench
        A ConbenchClient instance. If not given, one will be created using the standard
        environment variables.

    Environment variables
    ---------------------
    BUILD_URL
        The URL of the build running this code. If provided, the GitHub Check will link
        to the build when there's an error in this workflow.
    GITHUB_APP_ID
        The ID of a GitHub App that has been set up according to this package's
        instructions and installed to your repo. Only required if a GitHubRepoClient is
        not provided.
    GITHUB_APP_PRIVATE_KEY
        The private key file contents of a GitHub App that has been set up according to
        this package's instructions and installed to your repo. Only required if
        GitHubRepoClient is not provided.
    CONBENCH_URL
        The URL of the Conbench server. Only required if a ConbenchClient is not
        provided.
    CONBENCH_EMAIL
        The email to use for Conbench login. Only required if a ConbenchClient is not
        provided and the server is private.
    CONBENCH_PASSWORD
        The password to use for Conbench login. Only required if a ConbenchClient is not
        provided and the server is private.

    Returns
    -------
    dict
        GitHub's details about the new check.
    """
    build_url = os.getenv("BUILD_URL")
    github = github or GitHubRepoClient(repo=repo)

    def update_check(status, title, summary, details, details_url):
        """Shortcut for updating the "Conbench performance report" check on the given
        SHA, with debug logging.
        """
        res = github.update_check(
            name="Conbench performance report",
            commit_sha=contender_sha,
            status=status,
            title=title,
            summary=summary,
            details=details,
            details_url=details_url,
        )
        log.debug(res)
        return res

    # mark the task as pending
    update_check(
        status=CheckStatus.IN_PROGRESS,
        title="Analyzing performance",
        summary=f"Analyzing `{contender_sha[:8]}` for regressions...",
        details=None,
        details_url=build_url,
    )

    # If anything above this line fails, we can't tell github that it failed.
    # If anything in here fails, we can!
    try:
        conbench = conbench or ConbenchClient()
        comparisons = get_comparison_to_baseline(
            conbench, contender_sha, z_score_threshold
        )
        regressions = benchmarks_with_z_regressions(comparisons)

        status = regression_check_status(comparisons)
        summary = regression_summary(comparisons, warn_if_baseline_isnt_parent)
        details = regression_details(comparisons)
        if any(comparison.has_errors for comparison in comparisons):
            title = "Some benchmarks had errors"
        else:
            title = f"Found {len(regressions)} regression(s)"
        # point to the homepage table filtered to runs of this commit
        url = f"{os.environ['CONBENCH_URL']}/?search={contender_sha}"

        return update_check(
            status=status,
            title=title,
            summary=summary,
            details=details,
            details_url=url,
        )

    except Exception as e:
        summary = _clean(
            """
            The CI build running the regression analysis failed. This does not
            necessarily mean this commit has benchmark regressions, but there is an
            error that must be resolved before we can find out.
            """
        )
        details = f"Error: `{repr(e)}`\n\nSee build link below."
        update_check(
            status=CheckStatus.NEUTRAL,
            title="Error when analyzing performance",
            summary=summary,
            details=details,
            details_url=build_url,
        )
        log.error(f"Updated status with error: {e}")
        raise



# Copyright (c) 2022, Voltron Data.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pathlib
from typing import List

import setuptools


def read_requirements_file(filepath: pathlib.Path) -> List[str]:
    """Parse a requirements.txt file into a list of package requirements"""
    with open(filepath, "r") as f:
        requirements = [
            line.strip() for line in f if line.strip() and not line.startswith("#")
        ]
    return requirements


pkg_root = pathlib.Path(__file__).parent

__version__ = ""
with open(pkg_root / "benchalerts" / "_version.py", "r") as f:
    exec(f.read())  # only overwrites the __version__ variable

with open(pkg_root / "README.md", "r") as f:
    long_description = f.read()

base_requirements = read_requirements_file(pkg_root / "requirements.txt")
dev_requirements = read_requirements_file(pkg_root / "requirements-dev.txt")


setuptools.setup(
    name="benchalerts",
    version=__version__,
    description="Automated alerting for conbench",
    long_description=long_description,
    long_description_content_type="text/markdown",
    packages=setuptools.find_packages(),
    entry_points={},
    classifiers=[
        "Programming Language :: Python :: 3.8",
        "License :: OSI Approved :: Apache Software License",
    ],
    python_requires=">=3.8",
    maintainer="Austin Dickey",
    maintainer_email="austin@voltrondata.com",
    url="https://github.com/conbench/benchalerts",
    install_requires=base_requirements,
    extras_require={"dev": dev_requirements},
)
